class UnsupportedService(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class FileTooLarge(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class InvalidRequest(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class InvalidSession(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class AccessDenied(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class UnexistentData(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class ActionNotAllowed(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class ServiceUnderMaintenance(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class MessageNeeded(Exception):
    def __init__(self, response: str):
        super().__init__(response)
        
class InvalidAccountOrPassword(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class AccountDisabled(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class EmailAlreadyTaken(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class AccountDoesNotExist(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class InvalidDevice(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class AccountLimitReached(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class CantFollowYourself(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class UserUnavailable(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class YouAreBanned(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class UserNotMemberOfCommunity(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class RequestRejected(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class ActivateAccount(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class CantLeaveCommunity(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class ReachedTitleLength(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class EmailFlaggedAsSpam(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class AccountDeleted(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class ReachedMaxTitles(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class VerificationRequired(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class UserBannedByTeamAmino(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class BadImage(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class RequestedNoLongerExists(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class InsufficientLevel(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class WallCommentingDisabled(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class CommunityNoLongerExists(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class InvalidCodeOrLink(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class CommunityNameAlreadyTaken(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class CommunityCreateLimitReached(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class CommunityDisabled(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class CommunityDeleted(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class TooManyInviteUsers(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class ChatInvitesDisabled(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class RemovedFromChat(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class UserNotJoined(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class LevelFiveRequiredToEnableProps(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class ChatViewOnly(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class ChatMessageTooBig(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class InviteCodeNotFound(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class AlreadyRequestedJoinCommunity(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class AlreadyCheckedIn(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class AlreadyUsedMonthlyRepair(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class AccountAlreadyRestored(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class IncorrectVerificationCode(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class NotOwnerOfChatBubble(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class NotEnoughCoins(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class AlreadyPlayedLottery(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class CannotSendCoins(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class AminoIDAlreadyChanged(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class InvalidAminoID(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class InvalidName(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class YouHaveBlockedThisUser(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class NoLongerExists(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class API_ERROR(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class BlockedByOrganizer(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class YouAreBlockedByThisUser(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class UserHasBeenDeleted(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class DataNoLongerExists(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class InternalServerError(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class InvalidEmail(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class InvalidPassword(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class WhoaCooldown(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class InvalidThemepack(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class InvalidVoiceNote(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class PostedTooRecently(Exception):
    def __init__(self, response: str):
        super().__init__(response)

class APIException(Exception):
    def __init__(self, response: dict):
        exception_map = {
            100: UnsupportedService,
            101: InternalServerError,
            102: FileTooLarge,
            103: InvalidRequest,
            104: InvalidRequest,
            105: InvalidSession,
            106: AccessDenied,
            107: UnexistentData,
            110: ActionNotAllowed,
            111: ServiceUnderMaintenance,
            113: MessageNeeded,
            200: InvalidAccountOrPassword,
            210: AccountDisabled,
            213: InvalidEmail,
            214: InvalidPassword,
            215: EmailAlreadyTaken,
            216: AccountDoesNotExist,
            218: InvalidDevice,
            219: AccountLimitReached,
            221: CantFollowYourself,
            225: UserUnavailable,
            229: YouAreBanned,
            230: UserNotMemberOfCommunity,
            235: RequestRejected,
            238: ActivateAccount,
            239: CantLeaveCommunity,
            240: ReachedTitleLength,
            241: EmailFlaggedAsSpam,
            245: UserHasBeenDeleted,
            246: AccountDeleted,
            262: ReachedMaxTitles,
            270: VerificationRequired,
            291: WhoaCooldown,
            293: UserBannedByTeamAmino,
            300: BadImage,
            313: InvalidThemepack,
            314: InvalidVoiceNote,
            500: RequestedNoLongerExists,
            503: PostedTooRecently,
            551: InsufficientLevel,
            603: YouAreBlockedByThisUser,
            604: YouHaveBlockedThisUser,
            606: BlockedByOrganizer,
            700: NoLongerExists,
            702: WallCommentingDisabled,
            801: CommunityNoLongerExists,
            802: InvalidCodeOrLink,
            805: CommunityNameAlreadyTaken,
            806: CommunityCreateLimitReached,
            814: CommunityDisabled,
            826: CommunityCreateLimitReached,
            833: CommunityDeleted,
            1600: DataNoLongerExists,
            1606: TooManyInviteUsers,
            1611: ChatInvitesDisabled,
            1612: RemovedFromChat,
            1613: UserNotJoined,
            1661: LevelFiveRequiredToEnableProps,
            1663: ChatViewOnly,
            1664: ChatMessageTooBig,
            1900: InviteCodeNotFound,
            2001: AlreadyRequestedJoinCommunity,
            2601: AlreadyCheckedIn,
            2611: AlreadyUsedMonthlyRepair,
            2800: AccountAlreadyRestored,
            3102: IncorrectVerificationCode,
            3905: NotOwnerOfChatBubble,
            4300: NotEnoughCoins,
            4400: AlreadyPlayedLottery,
            4500: CannotSendCoins,
            6001: AminoIDAlreadyChanged,
            6002: InvalidAminoID,
            99001: InvalidName
        }

        status_code = response.get("api:statuscode")
        message = response.get("api:message")
        url = response.get("url")

        if status_code in exception_map:
            exception = exception_map.get(status_code)

            if url is not None:
                message = f"{message} ({url})"

            if exception:
                raise exception(message)

        raise API_ERROR(response)

class AccountLoginRatelimited(Exception):
    def __init__(self):
        super().__init__(
            "Account login ratelimited by Amino, wait 3 minutes and try again."
        )
            
class MissingCommunityId(Exception):
    def __init__(self):
        super().__init__(
            "Please provide a community id to the bot before running it or add it to the function call."
            )

class VerifyCommunityIdIsCorrect(Exception):
    def __init__(self):
        super().__init__(
            "Check your community id! It should be an integer.\nIf you're using a community link, use `fetch_community_id` instead."
            )
        
class MissingEmailPasswordOrSid(Exception):
    def __init__(self):
        super().__init__(
            "Please provide an email, password or sid to the bot.run() function before running it."
            )
        
class LoginFailed(Exception):
    def __init__(self):
        super().__init__(
            "Login failed. Please check your email and password."
            )
        
class FailedToFetchWebsocketUrl(Exception):
    def __init__(self):
        super().__init__(
            "Failed to fetch websocket url. Please try again later."
            )
        
class LoginRequired(Exception):
    def __init__(self):
        super().__init__(
            "Please login before using this function."
            )
        
class InvalidImage(Exception):
    def __init__(self):
        super().__init__(
            "Invalid image. Please check the image you are trying to upload. It should be a file path or a url."
            )
        
class MissingTimers(Exception):
    def __init__(self):
        super().__init__(
            "You are missing the start and end timers. Please provide them."
            )
        
class NoDataProvided(Exception):
    def __init__(self):
        super().__init__(
            "No data provided. Please provide data to the function."
            )
        
class NotLoggedIn(Exception):
    def __init__(self):
        super().__init__(
            "You are not logged in. Please login before using this function."
            )
        
class Forbidden(Exception):
    def __init__(self):
        super().__init__(
            "403 Forbidden. Possible IP ban. Please try again later."
            "\nIf this error persists, change your IP address.\n"
            )
        
class ServiceUnavailable(Exception):
    def __init__(self):
        super().__init__(
            "Amino is currently down or having issues. Please try again later."
            "\nThis has nothing to do with the bot or your internet connection.\n"
            )
        
class BadGateway(Exception):
    def __init__(self):
        super().__init__(
            "Amino is currently down or having issues. Please try again later."
            "\nThis has nothing to do with the bot or your internet connection.\n"
            )
        
class InvalidLink(Exception):
    def __init__(self):
        super().__init__(
            "Invalid link. Please check the link you are trying to use."
            )
        
class MustRunInContext(Exception):
    def __init__(self):
        super().__init__(
            "This function must be run in a context."
            )
        
class WrongWebSocketPackage(Exception):
    def __init__(self):
        super().__init__(
            "Wrong websocket package was installed. We corrected it for you."
            "\nPlease restart your bot.\n"
            )
        
class NullResponse(Exception):
    def __init__(self):
        super().__init__(
            "Server returned a null response. Possible crash content in the requested data."
            )
        
class PingFailed(Exception):
    def __init__(self):
        super().__init__(
            "Ping failed. Please make sure you are logged in and try again."
            )
        
class IntentsNotEnabled(Exception):
    def __init__(self):
        super().__init__(
            "Intents are not enabled. Please enable them in your Bot instance and try again."
            )

class InvalidCommandPrefix(Exception):
    def __init__(self):
        super().__init__(
            "Invalid command prefix. Please provide a valid command prefix."
            )
        
class MissingAwaitError(Exception):
    def __init__(self):
        super().__init__(
            "Missing await error. Please add `await` before your function call."
            )
        
class MissingDeviceKeyOrSignatureKey(Exception):
    def __init__(self):
        super().__init__(
            "Missing signature and or device key. Please provide a signature or device key."
            )

class MissingServiceKey(Exception):
    def __init__(self):
        super().__init__(
            "It appears you are missing a required service key. This key is essential for accessing the service. To obtain the key, please visit the official Discord server. The Server provides support and resources related to the service. https://discord.gg/3HRdkVNets"
        )


class EventTypes:
    def __init__(self) -> None:
        self.events = {
            "0:0": "text_message",
            "0:100": "image_message",
            "0:103": "youtube_message",
            "1:0": "strike_message",
            "2:110": "voice_message",
            "3:113": "sticker_message",
            "50:0": "share_exurl_message",
            "52:0": "vc_not_answered",
            "53:0": "vc_not_cancelled",
            "54:0": "vc_not_declined",
            "55:0": "video_chat_not_answered",
            "56:0": "video_chat_not_cancelled",
            "57:0": "video_chat_not_declined",
            "58:0": "avatar_chat_not_answered",
            "59:0": "avatar_chat_not_cancelled",
            "60:0": "avatar_chat_not_declined",
            "100:0": "delete_message",
            "101:0": "member_join",
            "102:0": "member_leave",
            "103:0": "chat_invite",
            "104:0": "chat_background_changed",
            "105:0": "chat_title_changed",
            "106:0": "chat_icon_changed",
            "107:0": "vc_start",
            "108:0": "video_chat_start",
            "109:0": "avatar_chat_start",
            "110:0": "vc_end",
            "111:0": "video_chat_end",
            "112:0": "avatar_chat_end",
            "113:0": "chat_content_changed",
            "114:0": "screen_room_start",
            "115:0": "screen_room_end",
            "116:0": "chat_host_transfered",
            "117:0": "text_message_force_removed",
            "118:0": "chat_removed_message",
            "119:0": "mod_deleted_message",
            "120:0": "chat_tip",
            "121:0": "chat_pin_announcement",
            "122:0": "vc_permission_open_to_everyone",
            "123:0": "vc_permission_invited_and_requested",
            "124:0": "vc_permission_invite_only",
            "125:0": "chat_view_only_enabled",
            "126:0": "chat_view_only_disabled",
            "127:0": "chat_unpin_announcement",
            "128:0": "chat_tipping_enabled",
            "129:0": "chat_tipping_disabled",
            "65281:0": "timestamp_message",
            "65282:0": "welcome_message",
            "65283:0": "invite_message"
            }
        
class NotifTypes:
    def __init__(self) -> None:
        self.notifs = {
        53: "member_set_you_host",
        67: "member_set_you_cohost",
        68: "member_remove_your_cohost"
        }


from typing import Union


class MemberAvatar:
    def __init__(self, data: dict) -> None:
        try:
            self.data = data
        except AttributeError:
            self.data = None
    
    def _check_member_avatar(F):
        def wrapper(*args, **kwargs):
            return None if args[0].data is None else F(*args, **kwargs)
        return wrapper
    
    @property
    @_check_member_avatar
    def status(self) -> Union[int, None]:
        """Returns the status of the member avatar."""
        return self.data.get("status")
    
    @property
    @_check_member_avatar
    def version(self) -> Union[int, None]:
        """Returns the version of the member avatar."""
        return self.data.get("version")
    
    @property
    @_check_member_avatar
    def resourceUrl(self) -> Union[str, None]:
        """Returns the resource URL of the member avatar."""
        return self.data.get("resourceUrl")
    
    @property
    @_check_member_avatar
    def name(self) -> Union[str, None]:
        """Returns the name of the member avatar."""
        return self.data.get("name")
    
    @property
    @_check_member_avatar
    def icon(self) -> Union[str, None]:
        """Returns the icon of the member avatar."""
        return self.data.get("icon")
    
    @property
    @_check_member_avatar
    def frameType(self) -> Union[int, None]:
        """Returns the frame type of the member avatar."""
        return self.data.get("frameType")
    
    @property
    @_check_member_avatar
    def frameId(self) -> Union[str, None]:
        """Returns the frame ID of the member avatar."""
        return self.data.get("frameId")


class Member:
    def __init__(self, data: dict) -> None:
        try:
            self.data = data
        except AttributeError:
            self.data = None
        
    def _check_member(F):
        def wrapper(*args, **kwargs):
            return None if args[0].data is None else F(*args, **kwargs)
        return wrapper
    
    @property
    @_check_member
    def uid(self) -> Union[str, None]:
        """Returns the uid of the member."""
        return self.data.get("uid")
    
    @property
    @_check_member
    def userId(self) -> Union[str, None]:
        """Returns the userId of the member."""
        return self.uid
    
    @property
    @_check_member
    def status(self) -> Union[int, None]:
        """Returns the status of the member."""
        return self.data.get("status")
    
    @property
    @_check_member
    def icon(self) -> Union[str, None]:
        """Returns the icon of the member."""
        return self.data.get("icon")
    
    @property
    @_check_member
    def reputation(self) -> Union[int, None]:
        """Returns the reputation of the member."""
        return self.data.get("reputation")
    
    @property
    @_check_member
    def role(self) -> Union[int, None]:
        """Returns the role of the member."""
        return self.data.get("role")
    
    @property
    @_check_member
    def nickname(self) -> Union[str, None]:
        """Returns the nickname of the member."""
        return self.data.get("nickname")
    
    @property
    @_check_member
    def level(self) -> Union[int, None]:
        """Returns the level of the member."""
        return self.data.get("level")
    
    @property
    @_check_member
    def membership_status(self) -> Union[int, None]:
        """Returns the accountMembershipStatus of the member."""
        return self.data.get("accountMembershipStatus")
    
    @property
    @_check_member
    def avatar(self) -> Union[MemberAvatar, None]:
        """Returns the avatar of the member."""
        return MemberAvatar(self.data.get("avatarFrame"))


from typing import List, Union


class Bubble:
    def __init__(self, data: dict) -> None:
        try:
            self.data = data
        except AttributeError:
            self.data = None
    
    def _check_bubble(F):
        def wrapper(*args, **kwargs):
            return None if args[0].data is None else F(*args, **kwargs)
        return wrapper

    @property
    @_check_bubble
    def ref_object_id(self) -> str:
        """Returns the ref object id of the bubble."""
        return self.data.get("refObjectId")

    @property
    @_check_bubble
    def created_time(self) -> str:
        """Returns the time the bubble was created."""
        return self.data.get("createdTime")

    @property
    @_check_bubble
    def item_basic_info(self) -> dict:
        """Returns the basic info of the bubble."""
        return self.data.get("itemBasicInfo")

    @property
    @_check_bubble
    def icon(self) -> str:
        """Returns the icon url of the bubble."""
        if self.item_basic_info is None:
            return None
        else:
            return self.item_basic_info.get("icon")
        
    @property
    @_check_bubble
    def name(self) -> str:
        """Returns the name of the bubble."""
        if self.item_basic_info is None:
            return None
        else:
            return self.item_basic_info.get("name")
        
    @property
    @_check_bubble
    def item_restriction_info(self) -> dict:
        """Returns the restriction info of the bubble."""
        return self.data.get("itemRestrictionInfo")


    @property
    @_check_bubble
    def owner_uid(self) -> Union[str, None]:
        """Returns the owner uid of the bubble."""
        if self.item_restriction_info is None:
            return None
        else:
            return self.item_restriction_info.get("ownerUid")
        
    @property
    @_check_bubble
    def owner_type(self) -> Union[int, None]:
        """Returns the owner type of the bubble."""
        if self.item_restriction_info is None:
            return None
        else:
            return self.item_restriction_info.get("ownerType")
        
    @property
    @_check_bubble
    def restrict_type(self) -> Union[int, None]:
        """Returns the restict type of the bubble."""
        if self.item_restriction_info is None:
            return None
        else:
            return self.item_restriction_info.get("restrictType")
        
    @property
    @_check_bubble
    def restrict_value(self) -> Union[int, None]:
        """Returns the restrict value of the bubble."""
        if self.item_restriction_info is None:
            return None
        else:
            return self.item_restriction_info.get("restrictValue")
        
    @property
    @_check_bubble
    def available_duration(self) -> Union[int, None]:
        """Returns the available duration of the bubble."""
        if self.item_restriction_info is None:
            return None
        else:
            return self.item_restriction_info.get("availableDuration")
        
    @property
    @_check_bubble
    def discount_value(self) -> Union[int, None]:
        """Returns the discount value of the bubble."""
        if self.item_restriction_info is None:
            return None
        else:
            return self.item_restriction_info.get("discountValue")
        
    @property
    @_check_bubble
    def discount_status(self) -> Union[int, None]:
        """Returns the discount status of the bubble."""
        if self.item_restriction_info is None:
            return None
        else:
            return self.item_restriction_info.get("discountStatus")
        
    @property
    @_check_bubble
    def ref_object(self) -> dict:
        """Returns the ref object of the bubble."""
        return self.data.get("refObject")

    @property
    @_check_bubble
    def is_global(self) -> bool:
        """Returns whether the bubble is globally available."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("isGloballyAvailable")
        
    @property
    @_check_bubble
    def available_community_ids(self) -> list:
        """Return the available community ids of the bubble."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("availableNdcIds")
        
    @property
    @_check_bubble
    def bubble_type(self) -> int:
        """Returns the bubble type of the bubble."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("bubbleType")
        
    @property
    @_check_bubble
    def bubble_id(self) -> str:
        """Returns the bubble id of the bubble."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("bubbleId")
        
    @property
    @_check_bubble
    def background_image(self) -> str:
        """Returns the background image of the bubble."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("backgroundImage")
        
    @property
    @_check_bubble
    def status(self) -> int:
        """Returns the status of the bubble."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("status")
        
    @property
    @_check_bubble
    def is_new(self) -> bool:
        """Returns whether the bubble is new."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("isNew")
        
    @property
    @_check_bubble
    def bubble_name(self) -> str:
        """Returns the name of the bubble."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("name")
        
    @property
    @_check_bubble
    def banner_image(self) -> str:
        """Returns the banner image of the bubble."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("bannerImage")
        
    @property
    @_check_bubble
    def resource_url(self) -> str:
        """Returns the resource url of the bubble."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("resourceUrl")
        
    @property
    @_check_bubble
    def ownership_status(self) -> dict:
        """Returns the ownership status of the bubble."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("ownershipStatus")
        
    @property
    @_check_bubble
    def deletable(self) -> bool:
        """Returns whether the bubble is deletable."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("deletable")
        
    @property
    @_check_bubble
    def config(self) -> dict:
        """Returns the config of the bubble."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("config")
        
    @property
    @_check_bubble
    def version(self) -> int:
        """Returns the version of the bubble."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("version")
        
    @property
    @_check_bubble
    def modified_time(self) -> str:
        """Returns the modified time of the bubble."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("modifiedTime")
        
    @property
    @_check_bubble
    def is_activated(self) -> bool:
        """Returns whether the bubble is activated."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("isActivated")
        
    @property
    @_check_bubble
    def cover_image(self) -> str:
        """Returns the cover image of the bubble."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("coverImage")
        
    @property
    @_check_bubble
    def extensions(self) -> dict:
        """Returns the extensions of the bubble."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("extensions")
        
    @property
    @_check_bubble
    def template_id(self) -> str:
        """Returns the template id of the bubble."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("templateId")

    @property
    @_check_bubble
    def uid(self) -> str:
        """Returns the uid of the bubble."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("uid")
        
    @property
    @_check_bubble
    def md5(self) -> str:
        """Returns the md5 of the bubble."""
        if self.ref_object is None:
            return None
        else:
            return self.ref_object.get("md5")
        
    @property
    @_check_bubble
    def ref_object_type(self) -> int:
        """Returns the ref object type of the bubble."""
        if self.data is None:
            return None
        else:
            return self.data.get("refObjectType")
        


class BubbleList:
    def __init__(self, data: dict):
        try:
            self.data = data.get("storeItemList")
        except Exception:
            self.data = None

    def _check_bubbles(F):
        def wrapper(*args, **kwargs):
            return None if args[0].data is None else F(*args, **kwargs)
        return wrapper
    
    def __iterator__(self):
        """
        Iterator function to iterate over the bubbles in the bubble list.
        """
        for bubble in self.data:
            yield Bubble(bubble)

    @property
    @_check_bubbles
    def ref_object_id(self) -> List[str]:
        """Returns the ref object id of the bubble list."""
        return [bubble.ref_object_id for bubble in self.__iterator__()]
    
    @property
    @_check_bubbles
    def created_time(self) -> List[str]:
        """Returns the created time of the bubble list."""
        return [bubble.created_time for bubble in self.__iterator__()]
    
    @property
    @_check_bubbles
    def item_basic_info(self) -> List[dict]:
        """Returns the item basic info of the bubble list."""
        return [bubble.item_basic_info for bubble in self.__iterator__()]
    
    @property
    @_check_bubbles
    def icon(self) -> List[str]:
        """Returns the icon of the bubble list."""
        return [bubble.icon for bubble in self.__iterator__()]
    
    @property
    @_check_bubbles
    def name(self) -> List[str]:
        """Returns the name of the bubble list."""
        return [bubble.name for bubble in self.__iterator__()]
    
    @property
    @_check_bubbles
    def item_restriction_info(self) -> List[dict]:
        """Returns the item restriction info of the bubble list."""
        return [bubble.item_restriction_info for bubble in self.__iterator__()]
    
    @property
    @_check_bubbles
    def owner_uid(self) -> List[str]:
        """Returns the owner uid of the bubble list."""
        return [bubble.owner_uid for bubble in self.__iterator__()]
    
    @property
    @_check_bubbles
    def owner_type(self) -> List[int]:
        """Returns the owner type of the bubble list."""
        return [bubble.owner_type for bubble in self.__iterator__()]
    
    @property
    @_check_bubbles
    def restrict_type(self) -> List[int]:
        """Returns the restrict type of the bubble list."""
        return [bubble.restrict_type for bubble in self.__iterator__()]
    
    @property
    @_check_bubbles
    def restrict_value(self) -> List[int]:
        """Returns the restrict value of the bubble list."""
        return [bubble.restrict_value for bubble in self.__iterator__()]
    
    @property
    @_check_bubbles
    def available_duration(self) -> List[str]:
        """Returns the available duration of the bubble list."""
        return [bubble.available_duration for bubble in self.__iterator__()]
    
    @property
    @_check_bubbles
    def discount_value(self) -> List[str]:
        """Returns the discount value of the bubble list."""
        return [bubble.discount_value for bubble in self.__iterator__()]
    
    @property
    @_check_bubbles
    def discount_status(self) -> List[int]:
        """Returns the discount status of the bubble list."""
        return [bubble.discount_status for bubble in self.__iterator__()]
    
    @property
    @_check_bubbles
    def ref_object(self) -> List[dict]:
        """Returns the ref object of the bubble list."""
        return [bubble.ref_object for bubble in self.__iterator__()]
    
    @property
    @_check_bubbles
    def is_global(self) -> List[str]:
        """Returns the is global of the bubble list."""
        return [bubble.is_global for bubble in self.__iterator__()]
    
    @property
    @_check_bubbles
    def available_community_ids(self) -> List[int]:
        """Returns the available community ids of the bubble list."""
        return [bubble.available_community_ids for bubble in self.__iterator__()]
    
    @property
    @_check_bubbles
    def bubble_type(self) -> List[int]:
        """Returns the bubble type of the bubble list."""
        return [bubble.bubble_type for bubble in self.__iterator__()]
    
    @property
    @_check_bubbles
    def bubble_id(self) -> List[str]:
        """Returns the bubble id of the bubble list."""
        return [bubble.bubble_id for bubble in self.__iterator__()]
    
    @property
    @_check_bubbles
    def background_image(self) -> List[str]:
        """Returns the background image of the bubble list."""
        return [bubble.background_image for bubble in self.__iterator__()]

    @property
    @_check_bubbles
    def status(self) -> List[int]:
        """Returns the status of the bubble list."""
        return [bubble.status for bubble in self.__iterator__()]

    @property
    @_check_bubbles
    def is_new(self) -> List[str]:
        """Returns the is new of the bubble list."""
        return [bubble.is_new for bubble in self.__iterator__()]

    @property
    @_check_bubbles
    def name(self) -> List[str]:
        """Returns the name of the bubble list."""
        return [bubble.name for bubble in self.__iterator__()]

    @property
    @_check_bubbles
    def banner_image(self) -> List[str]:
        """Returns the banner image of the bubble list."""
        return [bubble.banner_image for bubble in self.__iterator__()]

    @property
    @_check_bubbles
    def resource_url(self) -> List[str]:
        """Returns the resource url of the bubble list."""
        return [bubble.resource_url for bubble in self.__iterator__()]

    @property
    @_check_bubbles
    def ownership_status(self) -> List[str]:
        """Returns the ownership status of the bubble list."""
        return [bubble.ownership_status for bubble in self.__iterator__()]

    @property
    @_check_bubbles
    def deletable(self) -> List[str]:
        """Returns the deletable of the bubble list."""
        return [bubble.deletable for bubble in self.__iterator__()]

    @property
    @_check_bubbles
    def config(self) -> List[dict]:
        """Returns the config of the bubble list."""
        return [bubble.config for bubble in self.__iterator__()]

    @property
    @_check_bubbles
    def version(self) -> List[int]:
        """Returns the version of the bubble list."""
        return [bubble.version for bubble in self.__iterator__()]

    @property
    @_check_bubbles
    def modified_time(self) -> List[str]:
        """Returns the modified time of the bubble list."""
        return [bubble.modified_time for bubble in self.__iterator__()]

    @property
    @_check_bubbles
    def is_activated(self) -> List[str]:
        """Returns the is activated of the bubble list."""
        return [bubble.is_activated for bubble in self.__iterator__()]

    @property
    @_check_bubbles
    def cover_image(self) -> List[str]:
        """Returns the cover image of the bubble list."""
        return [bubble.cover_image for bubble in self.__iterator__()]

    @property
    @_check_bubbles
    def extensions(self) -> List[str]:
        """Returns the extensions of the bubble list."""
        return [bubble.extensions for bubble in self.__iterator__()]

    @property
    @_check_bubbles
    def template_id(self) -> List[str]:
        """Returns the template id of the bubble list."""
        return [bubble.template_id for bubble in self.__iterator__()]

    @property
    @_check_bubbles
    def uid(self) -> List[str]:
        """Returns the uid of the bubble list."""
        return [bubble.uid for bubble in self.__iterator__()]

    @property
    @_check_bubbles
    def md5(self) -> List[str]:
        """Returns the md5 of the bubble list."""
        return [bubble.md5 for bubble in self.__iterator__()]
    

    def json(self) -> dict:
        """Returns the json of the bubble list."""
        return self.data


from base64 import urlsafe_b64decode
from random import randint
from time import time, sleep
from diskcache import Cache
from json import loads
from re import search
from asyncio import sleep as asleep
from datetime import datetime
from os import system, environ
from pip import main as pipmain
from contextlib import suppress

from colorama import Fore, Style

CACHE_NAME = "cache"

def check_debugger() -> bool:
    """
    Checks if the program is being run in a debugger.
    """
    with suppress(Exception):
        return any([
            search("vscode", environ.get("TERM_PROGRAM")),
            search("pycharm", environ.get("TERM_PROGRAM")),
            is_repl(), is_android()
            ])

def install_wsaccel() -> None:
    """
    Try to install wsaccel if it isn't installed.
    """
    with Cache(CACHE_NAME) as cache:
        if cache.get("wsaccel"):
            return None

        try:
            from wsaccel import __version__
            cache.set("wsaccel", True)
            return True
        except ImportError:
            pipmain(["install", "wsaccel"])
            cache.set("wsaccel", True)
            system("cls || clear")
            return None
    
def orjson_exists() -> bool:
    """
    Checks if orjson is installed. If it isn't, it will install it.
    """
    return False

def is_android() -> bool:
    """
    Checks if the program is running on an Android device.
    """
    return any(key in environ for key in ("ANDROID_ROOT", "ANDROID_DATA"))

def is_repl() -> bool:
    """
    Checks if the program is running on a repl.it instance.
    """
    return any(key for key in environ if key.lower().startswith("repl"))

def notify() -> None:
    """
    Notifies the user that the bot is online.
    """
    sleep(.1)
    print(f"{Fore.RED}[!] {Fore.YELLOW}If you see this message, you can safely ignore it. The bot is still running and will continue to run until you stop it.{Style.RESET_ALL}\n")
    print(f"{Fore.RED}[!] {Fore.YELLOW}Press {Fore.RED}CTRL+C {Fore.YELLOW}to stop the bot.{Style.RESET_ALL}\n")

def parse_auid(sid: str) -> str:
    """Parses the user ID from a session ID."""
    decoded_sid = urlsafe_b64decode(f"{sid}==")
    decoded_json: dict = loads(decoded_sid[1:-20].decode())
    return decoded_json["2"]

def cache_login(email: str, device: str, sid: str):
    """Cache the login credentials for the current user."""
    with suppress(Exception):
        cache = Cache(CACHE_NAME)
        cache.set(email, {"device": device, "sid": sid}, expire=86400)

def fetch_cache(email: str) -> tuple:
    """Fetch the login credentials for the current user."""
    with suppress(Exception):
        cache = Cache(CACHE_NAME)
        cache.expire()
        return cache[email]["sid"], cache[email]["device"]

def cache_exists(email: str) -> bool:
    """Check if the cache exists for the current user."""
    with suppress(Exception):
        cache = Cache(CACHE_NAME)
        cache.expire()
        return email in cache
    
async def alive_loop(ws) -> None:
    run_check = any([is_android(), is_repl()])

    start_time          = time()
    last_activity_time  = start_time
    last_message_time   = start_time
    while True:
        current_time = time()
        notify() if run_check else None

        with suppress(Exception):

            if await ws._is_interval_elapsed(
                last_time=ws._last_pinged,
                interval=15
            ) and not ws.reconnecting:
                await ws.reconnect()
                
            if await ws._is_interval_elapsed(
                last_time=last_message_time,
                interval=10
                ):
                last_message_time = current_time
                ws.loop.create_task(ws._send_message())

            if current_time - start_time >= 86400:
                start_time = current_time

            if all([await ws._is_interval_elapsed(
                last_time=last_activity_time,
                interval=300
                ),
                current_time - start_time <= 36000
                ]):
                ws.loop.create_task(ws._activity_status())
                last_activity_time = current_time

            await asleep(0)


def run_alive_loop(ws) -> None:
    run_check = False if ws.console_enabled else any([is_android(), is_repl()])
    
    start_time          = time()
    last_activity_time  = start_time
    last_message_time   = start_time

    notify()
    while True:
        current_time = time()

        

        with suppress(Exception):

            if ws._last_message(last_message_time):
                ws._send_message()
                last_message_time = current_time

            if current_time - start_time >= 86400:
                start_time = current_time

            if all([ws._last_active(last_activity_time), current_time - start_time <= 36000]):
                ws._activity_status()
                last_activity_time = current_time

            sleep(randint(25, 50))


from time import time

from .utilities.generate import Generator
from .entities.userprofile import UserProfile
from .entities.general import (
    ApiResponse, Authenticate, ResetPassword, Wallet
    )

class Account:
    """
    Account class for handling account related requests.
    """
    def __init__(self, session):
        self.session = session

    def register(self, email: str, password: str, username: str, verificationCode: str) -> Authenticate:
        """
        `**register**` - Registers a new account.

        `**Parameters**`

        - `email` - The email of the account.

        - `password` - The password of the account.

        - `username` - The username of the account.

        - `verificationCode` - The verification code sent to the email.
        
        `**Example**`

        ```py
        from pymino import *

        bot = Bot()
        bot.request_security_validation(email=email)
        code = input("Enter the code you received: ")
        response = bot.register(email=email, password=password, username=username, verificationCode=code)
        print(response.json())
        ```
        """
        return Authenticate(self.session.handler(
            method = "POST",
            url="/g/s/auth/register",
            data={
                "secret": f"0 {password}",
                "deviceID": self.session.generate.device_id(),
                "email": email,
                "clientType": 100,
                "nickname": username,
                "validationContext": {
                    "data": {"code": verificationCode}, "type": 1, "identity": email},
                "type": 1,
                "identity": email,
                "timestamp": int(time() * 1000)
                }))

    def delete_request(self, email: str, password: str) -> ApiResponse:
        """
        `**delete_request**` - Sends a delete request to the account.

        `**Parameters**`

        - `email` - The email of the account.

        - `password` - The password of the account.
        
        `**Example**`

        ```py
        from pymino import *
        
        bot = Bot()
        bot.run(email=email, password=password)
        response = bot.delete_request(email=email, password=password)
        print(response)
        ```"""
        return ApiResponse(self.session.handler(
            method = "POST",
            url="/g/s/account/delete-request",
            data={
                "secret": f"0 {password}",
                "deviceID": self.session.generate.device_id(),
                "email": email,
                "timestamp": int(time() * 1000)
            }))

    def delete_request_cancel(self, email: str, password: str) -> ApiResponse:
        """
        `**delete_request_cancel**` - Cancels the delete request.

        `**Parameters**`

        - `email` - The email of the account.

        - `password` - The password of the account.
        
        `**Example**`
        
        ```py
        from pymino import *
        
        bot = Bot()
        bot.run(email=email, password=password)
        response = bot.delete_request_cancel(email=email, password=password)
        print(response)
        ```
        """
        return ApiResponse(self.session.handler(
            method = "POST",
            url="/g/s/account/delete-request/cancel",
            data={
                "secret": f"0 {password}",
                "deviceID": self.session.generate.device_id(),
                "email": email,
                "timestamp": int(time() * 1000)
            }))

    def check_device(self, deviceId: str) -> ApiResponse:
        """
        `**check_device**` - Checks if the device is valid.

        `**Parameters**`

        - `deviceId` - The device id of the account.
        
        `**Example**`
        
        ```py
        from pymino import *
        
        bot = Bot()
        response = bot.check_device(deviceId=device_id())
        print(response)
        ```
        """
        return ApiResponse(self.session.handler(
            method = "POST",
            url="/g/s/device",
            data={
                "deviceID": deviceId,
                "clientType": 100,
                "timezone": -310,
                "systemPushEnabled": True,
                "timestamp": int(time() * 1000)
                }))

    def fetch_account(self) -> ApiResponse:
        """
        `**fetch_account**` - Fetches the account information.
        
        `**Example**`
        
        ```py
        from pymino import *
        
        bot = Bot()
        bot.run(email=email, password=password)
        response = bot.fetch_account()
        print(response)
        ```
        """
        return ApiResponse(self.session.handler(method = "GET", url="/g/s/account"))

    def upload_image(self, image: str) -> str:
        """
        `**upload_image**` - Uploads an image to the server.

        `**Parameters**`

        - `image` - The image to upload.
        
        `**Example**`
        
        ```py
        from pymino import *
        
        bot = Bot()
        bot.run(email=email, password=password)
        response = bot.upload_image(image="image.jpg")
        print(response)
        ```
        """
        return ApiResponse(self.session.handler(method="POST", url="/g/s/media/upload",
            data=open(image, "rb").read(), content_type="image/jpg")).mediaValue

    def fetch_profile(self, userId: str) -> UserProfile:
        """
        `**fetch_profile**` - Fetches the profile information.
        
        `**Example**`
        
        ```py
        from pymino import *
        
        bot = Bot()
        bot.run(email=email, password=password)
        response = bot.fetch_profile()
        print(response)
        ```
        """
        return UserProfile(self.session.handler(
            method = "GET", url = f"/g/s/user-profile/{userId}"))

    def set_amino_id(self, aminoId: str) -> ApiResponse:
        """
        `**set_amino_id**` - Sets the amino id.

        `**Parameters**`

        - `aminoId` - The amino id to set.
        
        `**Example**`
        
        ```py
        from pymino import *
        
        bot = Bot()
        bot.run(email=email, password=password)
        response = bot.set_amino_id(aminoId="aminoId")
        print(response)
        ```
        """
        return ApiResponse(self.session.handler(
            method="POST",
            url="/g/s/account/change-amino-id",
            data={"aminoId": aminoId, "timestamp": int(time() * 1000)}))

    def fetch_wallet(self) -> Wallet:
        """
        `**fetch_wallet**` - Fetches the wallet information.
        
        `**Example**`
        
        ```py
        from pymino import *
        
        bot = Bot()
        bot.run(email=email, password=password)
        response = bot.fetch_wallet()
        print(response)
        """
        return Wallet(self.session.handler(method="GET", url="/g/s/wallet"))

    def request_security_validation(self, email: str, resetPassword: bool = False) -> ApiResponse:
        """
        `**request_security_validation**` - Requests a security validation.

        `**Parameters**`

        - `email` - The email of the account.

        - `resetPassword` - Whether to reset the password or not.
        
        `**Example**`
        
        ```py
        from pymino import *
        
        bot = Bot()
        response = bot.request_security_validation(email=email)
        print(response)
        ```
        """
        return ApiResponse(self.session.handler(
            method = "POST", url="/g/s/auth/request-security-validation",
            data={
                "identity": email,
                "type": 1,
                "deviceID": self.session.generate.device_id(),
                "level": 2 if resetPassword else None,
                "purpose": "reset-password" if resetPassword else None,
                "timestamp": int(time() * 1000)
            }))

    def activate_email(self, email: str, code: str) -> ApiResponse:
        """
        `**activate_email**` - Activates an email.

        `**Parameters**`

        - `email` - The email of the account.

        - `code` - The code sent to the email.
        
        `**Example**`
        
        ```py
        from pymino import *
        
        bot = Bot()
        response = bot.activate_email(email=email, code=code)
        print(response)
        ```
        """
        return ApiResponse(self.session.handler(
            method = "POST",
            url="/g/s/auth/activate-email",
            data={
                "type": 1,
                "identity": email,
                "data": {"code":code},
                "deviceID": self.session.generate.device_id(),
                "timestamp": int(time() * 1000)
            }))
    
    
    def verify(self, email: str, code: str, deviceId: str) -> ApiResponse:
        """
        `**verify**` - Verifies the code sent to the email.

        `**Parameters**`
        
        - `email` - The email of the account.
        
        - `code` - The code sent to the email.

        - `deviceId` - The device id.
        
        `**Example**`
        
        ```py
        from pymino import *
        
        bot = Bot()
        response = bot.verify(email=email, code=code, deviceId=deviceId)
        print(response)
        ```
        """
        return ApiResponse(self.session.handler(
            method = "POST",
            url="/g/s/auth/check-security-validation",
            data={
                "type": 1,
                "identity": email,
                "data": {"code":code},
                "deviceID": deviceId,
                "timestamp": int(time() * 1000)
            }))

    def reset_password(self, email: str, newPassword: str, code: str, deviceId: str) -> ResetPassword:
        """
        `**reset_password**` - Resets the password.

        `**Parameters**`

        - `email` - The email of the account.

        - `newPassword` - The new password of the account.

        - `code` - The code sent to the email.
        
        `**Example**`
        
        ```py
        from pymino import *
        
        bot = Bot()
        bot.run(email=email, password=password)
        response = bot.reset_password(email=email, newPassword=newPassword, code=code)
        print(response)
        ```
        """
        return ResetPassword(self.session.handler(
            method = "POST", url="/g/s/auth/reset-password",
            data={
                "updateSecret": f"0 {newPassword}",
                "emailValidationContext": {
                    "data": {"code": code},
                    "type": 1,
                    "identity": email,
                    "level": 2,
                    "deviceID": deviceId
                },
                "phoneNumberValidationContext": None,
                "deviceID": deviceId
            }))


class CThreadExtensions:
    def __init__(self, data: dict):
        self.data = data
        parser = self.__parser__ if isinstance(self.data, list) else lambda key, default=None: self.data.get(key, default)
        self.coHost:                        list = parser('coHost', None)
        self.language:                      str = parser('language', None)
        self.membersCanInvite:              int = parser('membersCanInvite', None)
        self.background:                    list = parser('bm', None)
        self.creatorUid:                    str = parser('creatorUid', None)
        self.visibility:                    int = parser('visibility', None)
        self.lastMembersSummaryUpdateTime:  dict = parser('lastMembersSummaryUpdateTime', None)
        self.fansOnly:                      bool = parser('fansOnly', None)
        self.channelType:                   int = parser('channelType', None)
        self.vvChatJoinType:                int = parser('vvChatJoinType', None)
        self.avchatMemberUidList:           list = parser('avchatMemberUidList', None)
        self.screeningRoomPermission:       dict = parser('screeningRoomPermission', None)
        self.disabledTime:                  int = parser('__disabledTime__', None)

    def __parser__(self, key: str, default=None): return [i.get(key, default) for i in self.data]

    def json(self): return self.data

class MemberSummary:
    def __init__(self, data: dict):
        try:
            self.data = data[0] if isinstance(data[0], list) else data
        except IndexError:
            self.data = data
            
        self.status:            list = [x.get("status") for x in self.data]
        self.uid:               list = [x.get("uid") for x in self.data]
        self.userId:            list = self.uid
        self.membershipStatus:  list = [x.get("membershipStatus") for x in self.data]
        self.role:              list = [x.get("role") for x in self.data]
        self.nickname:          list = [x.get("nickname") for x in self.data]
        self.username:          list = self.nickname
        self.icon:              list = [x.get("icon") for x in self.data]
        self.avatar:            list = self.icon

    def json(self): return self.data

class CThread: 
    def __init__(self, data: dict):
        self.data:                  dict = data.get("thread", data)
        self.userAddedTopicList:    list = self.data.get("userAddedTopicList")
        self.uid:                   str = self.data.get("uid")
        self.hostUserId:            str = self.uid
        self.membersQuota:          int = self.data.get("membersQuota")
        self.membersSummary:        MemberSummary = MemberSummary(self.data.get("membersSummary"))
        self.threadId:              str = self.data.get("threadId")
        self.chatId:                str = self.threadId
        self.keywords:              list = self.data.get("keywords")
        self.membersCount:          int = self.data.get("membersCount")
        self.strategyInfo:          dict = self.data.get("strategyInfo")
        self.isPinned:              bool = self.data.get("isPinned")
        self.title:                 str = self.data.get("title")
        self.membershipStatus:      str = self.data.get("membershipStatus")
        self.content:               str = self.data.get("content")
        self.needHidden:            bool = self.data.get("needHidden")
        self.alertOption:           int = self.data.get("alertOption")
        self.lastReadTime:          str = self.data.get("lastReadTime")
        self.type:                  int = self.data.get("type")
        self.status:                int = self.data.get("status")
        self.publishToGlobal:       bool = self.data.get("publishToGlobal")
        self.modifiedTime:          str = self.data.get("modifiedTime")
        self.lastMessageSummary:    dict = self.data.get("lastMessageSummary")
        self.extensions:            CThreadExtensions = CThreadExtensions(self.data.get("extensions"))

    def json(self): return self.data

class CThreadList:
    def __init__(self, data: dict):
        self.data:                  dict = data.get("threadList", data)
        parser:                     list = [CThread(x) for x in self.data]
        self.extensions:            CThreadExtensions = CThreadExtensions([x.extensions.json() for x in parser])
        self.membersSummary:        MemberSummary = MemberSummary([x.membersSummary.json() for x in parser])
        self.userAddedTopicList:    list = [x.userAddedTopicList for x in parser]
        self.uid:                   list = [x.uid for x in parser]
        self.hostUserId:            list = [x.hostUserId for x in parser]
        self.membersQuota:          list = [x.membersQuota for x in parser]
        self.threadId:              list = [x.threadId for x in parser]
        self.chatId:                list = [x.chatId for x in parser]
        self.keywords:              list = [x.keywords for x in parser]
        self.membersCount:          list = [x.membersCount for x in parser]
        self.strategyInfo:          list = [x.strategyInfo for x in parser]
        self.isPinned:              list = [x.isPinned for x in parser]
        self.title:                 list = [x.title for x in parser]
        self.membershipStatus:      list = [x.membershipStatus for x in parser]
        self.content:               list = [x.content for x in parser]
        self.needHidden:            list = [x.needHidden for x in parser]
        self.alertOption:           list = [x.alertOption for x in parser]
        self.lastReadTime:          list = [x.lastReadTime for x in parser]
        self.type:                  list = [x.type for x in parser]
        self.status:                list = [x.status for x in parser]
        self.publishToGlobal:       list = [x.publishToGlobal for x in parser]
        self.modifiedTime:          list = [x.modifiedTime for x in parser]
        self.lastMessageSummary:    list = [x.lastMessageSummary for x in parser]

    def json(self): return self.data



from time import time
from typing import Callable
from collections import defaultdict


class Command:
    """
    `Command` - The main command class.
    
    `**Parameters**`
    - `func` - The function to run when the command is called.
    - `name` - The name of the command.
    - `description` - The description of the command. `Defaults` to `None`.
    - `usage` - The usage of the command. `Defaults` to `None`.
    - `aliases` - The aliases of the command. `Defaults` to `None`.
    - `cooldown` - The cooldown of the command. `Defaults` to `0`.
    
    """
    def __init__(self, func: Callable, name: str, description: str=None, usage: str=None, aliases: list=None, cooldown: int=0):
        self.func:           Callable = func
        self.name:           str = name
        self.description:    str = description
        self.usage:          str = usage
        self.aliases:        list = [] if aliases is None else aliases
        self.cooldown:       int = cooldown


class Commands:
    def __init__(self):
        self.commands: dict[str, Command] = {}
        self.cooldowns: defaultdict[dict] = defaultdict(dict)


    def add_command(self, command: Command) -> Command:
        """
        `add_command` - Adds a command to the command list.
        
        `**Parameters**`
        - `command` - The command to add.
        
        `**Returns**`
        - `Command` - The command that was added.
        
        """
        self.commands[command.name] = command
        return command


    def fetch_command(self, command_name: str) -> Command:
        """
        `fetch_command` - Fetches a command from the command list.
        
        `**Parameters**`
        - `command_name` - The name of the command to fetch.
        
        `**Returns**`
        - `Command` - The command that was fetched.
        
        """
        return self.commands.get(command_name) or next(
            (command for command in self.commands.values() if command_name in command.aliases), None
        )


    def fetch_commands(self) -> Command:
        """
        `fetch_commands` - Fetches all commands from the command list.
        
        `**Returns**`
        - `list[Command]` - The commands that were fetched.
        
        """
        return list(self.commands.values())


    def __command_functions__(self) -> list:
        """
        `__command_functions__` - Fetches all command functions from the command list.
        
        `**Returns**`
        - `list[Callable]` - The command functions that were fetched.
        
        """
        return {command.name: command.func for command in self.commands}


    def __command_names__(self) -> list:
        """
        `__command_names__` - Fetches all command names from the command list.
        
        `**Returns**`
        - `list[str]` - The command names that were fetched.
        
        """
        return [command.name for command in self.commands.values()]


    def __command_aliases__(self) -> list:
        """
        `__command_aliases__` - Fetches all command aliases from the command list.
        
        `**Returns**`
        - `list[str]` - The command aliases that were fetched.
        
        """
        aliases = {}
        for command in self.commands.values():
            for alias in command.aliases:
                aliases[alias] = command.name
        return aliases


    def __command_descriptions__(self) -> list:
        """
        `__command_descriptions__` - Fetches all command descriptions from the command list.
        
        `**Returns**`
        - `list[str]` - The command descriptions that were fetched.
        
        """
        return {command.name: command.description for command in self.commands}


    def __command_usages__(self) -> list:
        """
        `__command_usages__` - Fetches all command usages from the command list.
        
        `**Returns**`
        - `list[str]` - The command usages that were fetched.
        
        """
        return {command.name: command.usage for command in self.commands}


    def __command_cooldowns__(self) -> list:
        """
        `__command_cooldowns__` - Fetches all command cooldowns from the command list.
        
        `**Returns**`
        - `list[int]` - The command cooldowns that were fetched.
        
        """
        return {command.name: command.cooldown for command in self.commands}


    def set_cooldown(self, command_name: str, cooldown: int, userId: str) -> None:
        """
        `set_cooldown` - Sets the cooldown of a command for a user.
        
        `**Parameters**`
        - `command_name` - The name of the command to set the cooldown for.
        - `cooldown` - The cooldown to set.
        - `userId` - The user to set the cooldown for.
        
        `**Returns**`
        - `None` - Nothing.

        """
        self.cooldowns[command_name][userId] = time() + cooldown


    def fetch_cooldown(self, command_name: str, userId: str) -> int:
        """
        `fetch_cooldown` - Fetches the cooldown of a command for a user.
        
        `**Parameters**`
        - `command_name` - The name of the command to fetch the cooldown for.
        - `userId` - The user to fetch the cooldown for.
        
        `**Returns**`
        - `int` - The cooldown that was fetched.
        
        """
        self.cooldowns[command_name].setdefault(userId, 0)
        return self.cooldowns[command_name][userId]


    def __help__(self):
        """
        `__help__` - Generates a help message for the bot.
        
        `**Returns**`
        - `str` - The help message that was generated.
        
        """
        help_message = "[bcu]Commands\n" + "\n[ic]This is a list of all the commands available on this bot.\n"

        for command in self.commands.values():
            help_message += f"\n[uc]{command.name}\n[ic]{command.description}\n[uc]Usage: {command.usage}\n"
        help_message += "\n\n[ic]This message was generated automatically. If you have any questions, please contact the bot owner."
        return help_message


from enum import IntEnum, Enum

class ObjectTypes(IntEnum):
    USER = 0
    BLOG = 1
    WIKI = 2
    CHAT = 12

class ChatTypes(Enum):
    LATEST = "latest"
    POPULAR = "popular"
    RECOMMENDED = "recommended"

class LeaderboardTypes(IntEnum):
    ACTIVE24H = 1
    ACTIVE7D = 2
    HALL_OF_FAME = 3
    CHECK_IN = 4
    QUIZ = 5

class UserTypes(Enum):
    LEADERS = "leaders"
    CURATORS = "curators"       
    RECENT = "recent"
    FEATURED = "featured"    
    BANNED = "banned"

class FlagTypes(IntEnum):
    AGGRESSION =  0
    SPAM = 2
    OFFTOPIC = 4
    VIOLENCE = 106
    INTOLERANCE = 107
    SUICIDE = 108
    TROLLING = 109
    PORNOGRAPHY = 110

class OnlineTypes(IntEnum):
    ONLINE = 1
    OFFLINE = 2

class BanTypes(IntEnum):
    SEXUALLY_EXPLICIT = 100
    SPAM = 2
    OUT_OF_CONTEXT = 4
    HARASSMENT = 0
    INNAPROPIATE_SOLICITUDE = 102
    EXTREME_VIOLENCE = 101
    OTHER = 200